"""CLI commands."""
from __future__ import annotations

from pyrgo.cli.cmds._audit import audit
from pyrgo.cli.cmds._build import build
from pyrgo.cli.cmds._check import check
from pyrgo.cli.cmds._clean import clean
from pyrgo.cli.cmds._fmt import fmt
from pyrgo.cli.cmds._lock import lock
from pyrgo.cli.cmds._new import new
from pyrgo.cli.cmds._sync import sync
from pyrgo.cli.cmds._test import test

__all__ = ["lock", "fmt", "check", "sync", "build", "clean", "test", "audit", "new"]



"""Pyrgo configuration."""
from __future__ import annotations

import pathlib
import sys
from dataclasses import dataclass
from typing import Any

import loguru
import tomli


class PyProjectNotFoundError(Exception):
    """Raised when `pyproject.toml` not found."""

    def __init__(self, path: pathlib.Path) -> None:  # noqa: D107
        super().__init__(f"`pyproject.toml` not found at {path.as_posix()}")


def _configure_logger(logger: loguru.Logger) -> loguru.Logger:
    logger.remove()
    fmt = "<lvl>[{level}]</lvl> {message} <green>{name}:{function}:{line}</green> @ {time:HH:mm:ss}"  # noqa: E501
    logger.add(sys.stderr, format=fmt)
    return logger


@dataclass(frozen=True)
class PyrgoConf:
    """Pyrgo configuration."""

    cwd: pathlib.Path
    requirements: pathlib.Path
    logger: loguru.Logger
    relevant_paths: list[str]
    artifacts: list[pathlib.Path]
    caches: list[pathlib.Path]
    core_deps_alias: str
    env_groups: set[str]

    @classmethod
    def new(cls: type[PyrgoConf]) -> PyrgoConf:
        """Create new configuration instance."""
        logger = _configure_logger(logger=loguru.logger)
        cwd = pathlib.Path().cwd()
        pyproject_path = cwd.joinpath("pyproject.toml")
        if not (pyproject_path.exists() and pyproject_path.is_file()):
            raise PyProjectNotFoundError(path=cwd)

        pyproject_data = tomli.loads(pyproject_path.read_text())
        relevant_paths: list[str] = [
            pyproject_data["project"]["name"].strip().replace("-", "_"),
        ]
        relevant_paths.extend(
            pyproject_data["tool"]["pytest"]["ini_options"]["testpaths"],
        )
        pyrgo_config = pyproject_data["tool"].get("pyrgo", None)
        caches = [
            cwd.joinpath(".pytest_cache"),
            cwd.joinpath(".ruff_cache"),
            cwd.joinpath(".mypy_cache"),
        ]
        if pyrgo_config is not None:
            extra_paths = pyrgo_config.get("extra-paths", None)
            extra_caches = pyrgo_config.get("extra-caches", None)
            if extra_paths is not None:
                relevant_paths.extend(extra_paths)
            if extra_caches is not None:
                caches.extend(cwd.joinpath(extra) for extra in extra_caches)

        core_deps_alias = "core"
        env_groups = {core_deps_alias}
        op_deps: dict[str, Any] | None = pyproject_data["project"].get(
            "optional-dependencies",
            None,
        )
        if op_deps is not None:
            env_groups = env_groups.union(op_deps.keys())

        return cls(
            cwd=cwd,
            requirements=cwd.joinpath("requirements"),
            logger=logger,
            relevant_paths=relevant_paths,
            artifacts=[cwd.joinpath("dist")],
            caches=caches,
            core_deps_alias=core_deps_alias,
            env_groups=env_groups,
        )

    def locked_envs(self) -> set[str]:
        """Get a set of available locked envs."""
        locked_envs: set[str] = set()
        for env in self.requirements.glob(pattern="*.txt"):
            locked_envs.add(
                env.relative_to(self.requirements).as_posix().removesuffix(".txt"),
            )

        return locked_envs



"""Test command."""
from __future__ import annotations

import sys

import click
from result import Ok

from pyrgo.cli.cmds._utils import inform_and_run_program
from pyrgo.command_exec import PythonCommandExec


@click.command("test")
@click.option("-m", "marker", type=click.STRING, default=None, show_default=True)
def test(marker: str | None) -> None:
    """Run tests with `pytest`."""
    pytest_command = PythonCommandExec.new(program="pytest")
    if marker is not None:
        pytest_command.add_args(args=["-m", marker])
    program_execution = inform_and_run_program(commands=[pytest_command])
    if not isinstance(program_execution, Ok):
        sys.exit(1)

    sys.exit(0)



"""Check command."""
from __future__ import annotations

import sys

import click
from result import Ok

from pyrgo.cli.cmds._utils import inform_and_run_program
from pyrgo.command_exec import PythonCommandExec
from pyrgo.conf import PyrgoConf


@click.command("check")
@click.option(
    "-t",
    "--timeout",
    "timeout",
    default=360,
    type=click.IntRange(min=60),
)
@click.option(
    "--add-noqa",
    "add_noqa",
    is_flag=True,
    default=False,
    show_default=True,
    type=click.BOOL,
)
@click.option(
    "--ignore-noqa",
    "ignore_noqa",
    is_flag=True,
    default=False,
    show_default=True,
    type=click.BOOL,
)
def check(*, timeout: int, add_noqa: bool, ignore_noqa: bool) -> None:
    """Check code with `mypy` and `ruff`."""
    configuration = PyrgoConf.new()
    ruff_command = PythonCommandExec.new(program="ruff")
    mypy_command = PythonCommandExec.new(program="mypy.dmypy").add_args(
        args=["run", "--timeout", str(timeout), "--"],
    )

    if add_noqa:
        ruff_command.add_args(args=["--add-noqa"])
    if ignore_noqa:
        ruff_command.add_args(args=["--ignore-noqa"])

    for command in [ruff_command, mypy_command]:
        command.add_args(args=configuration.relevant_paths)

    program_execution = inform_and_run_program(commands=[ruff_command, mypy_command])
    if not isinstance(program_execution, Ok):
        sys.exit(1)

    sys.exit(0)



"""Command execution module."""
from __future__ import annotations

import subprocess
import sys
from dataclasses import dataclass
from typing import TYPE_CHECKING, Literal

from result import Err, Ok, Result

if TYPE_CHECKING:
    from typing_extensions import Self, TypeAlias


PyrgoProgram: TypeAlias = Literal[
    "ruff",
    "black",
    "mypy.dmypy",
    "piptools",
    "pip",
    "build",
    "pytest",
    "pip_audit",
]


@dataclass(frozen=False)
class PythonCommandExec:
    """Python command executor."""

    args: list[str]

    @classmethod
    def new(cls: type[PythonCommandExec], program: PyrgoProgram) -> PythonCommandExec:
        """Build a new command executor."""
        return cls(args=[sys.executable, "-m", program])

    def add_args(self, args: list[str]) -> Self:
        """Add arguments to command."""
        self.args.extend(args)
        return self

    def execute(self) -> Result[None, subprocess.CalledProcessError]:
        """Execute python command."""
        try:
            subprocess.run(args=self.args, check=True)
        except subprocess.CalledProcessError as e:
            return Err(e)
        return Ok(None)



"""CLI entrypoint."""
from __future__ import annotations

import click

from pyrgo.cli import cmds


@click.group(
    context_settings={
        "help_option_names": [
            "-h",
            "--help",
        ],
        "show_default": True,
    },
)
@click.version_option(None, "-v", "--version")
def cli() -> None:
    """Pyrgo. Python package manager."""


cli.add_command(cmd=cmds.lock)
cli.add_command(cmd=cmds.fmt)
cli.add_command(cmd=cmds.check)
cli.add_command(cmd=cmds.sync)
cli.add_command(cmd=cmds.build)
cli.add_command(cmd=cmds.clean)
cli.add_command(cmd=cmds.test)
cli.add_command(cmd=cmds.audit)
cli.add_command(cmd=cmds.new)



"""Format command."""
from __future__ import annotations

import sys

import click
from result import Ok

from pyrgo.cli.cmds._utils import inform_and_run_program
from pyrgo.command_exec import PythonCommandExec
from pyrgo.conf import PyrgoConf


@click.command("fmt")
def fmt() -> None:
    """Format code with `ruff` and `black`."""
    configuration = PyrgoConf.new()
    ruff_command = PythonCommandExec.new(program="ruff").add_args(args=["--fix-only"])
    black_command = PythonCommandExec.new(program="black")

    for command in [ruff_command, black_command]:
        command.add_args(args=configuration.relevant_paths)

    program_execution = inform_and_run_program(commands=[ruff_command, black_command])
    if not isinstance(program_execution, Ok):
        sys.exit(1)
    sys.exit(0)



"""CLI utilities."""
from __future__ import annotations

import sys
from typing import TYPE_CHECKING

import click
from result import Err, Ok, Result

if TYPE_CHECKING:
    import subprocess

    from pyrgo.command_exec import PythonCommandExec
    from pyrgo.conf import PyrgoConf


def inform_and_run_program(
    commands: list[PythonCommandExec],
) -> Result[None, list[subprocess.CalledProcessError]]:
    """Inform and execute command."""
    subprocess_errors: list[subprocess.CalledProcessError] = []
    for command in commands:
        command_to_execute = " ".join(command.args)
        click.echo(message=click.style(command_to_execute, fg="yellow"), color=True)
        execution_result = command.execute()
        if not isinstance(execution_result, Ok):
            subprocess_errors.append(execution_result.err())

    if len(subprocess_errors) > 0:
        return Err(subprocess_errors)
    return Ok(None)


def ensure_env_exist_in_lock_file(env: str, config: PyrgoConf) -> None:
    if not config.requirements.exists():
        click.echo(
            message=click.style(
                "No `requirements` folder found. Try locking your deps first",
                fg="red",
            ),
        )
        sys.exit(1)

    locked_envs = config.locked_envs()
    if env not in locked_envs:
        click.echo(
            click.style(
                f"`{env}` not found in available envs.\nAvailable envs: {locked_envs}",  # noqa: E501
                fg="red",
            ),
        )

        sys.exit(1)
